/*
 * Functions for Mac Input methods using Carbon
 *
 * Copyright 2006 CodeWeavers, Aric Stewart
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * 
 * Much of this code based off of Kinput2.
 *
 * Copyright (c) 1990  Software Research Associates, Inc.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  Software Research
 * Associates makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * Author:  Makoto Ishisone, Software Research Associates, Inc., Japan
 *
 */

/*
 * Copyright 1998 Takanori Sonoda. 
 *
 * Author: Takanori Sonoda, Kyushu Univ.  (tsonoda@gynob.med.kyushu-u.ac.jp)
 */

/*
 * Copyright (c) 2004 kenichi kikuchi
 * Author: kenichi kikuchi
 */

#ifdef __APPLE__

#include "wine/debug.h"
#include <Carbon/Carbon.h>

WINE_DEFAULT_DEBUG_CHANNEL(mac_xim);

#define kUniCharReverseSolidus	0x005c
#define kUniCharTilde			0x007e
#define kUniCharYenSign			0x00a5
#define kUniCharOverline		0x203e
#define kCandWindWidth	(127 + 11 + 80 + 11)
#define kMenuBarHeight	22

static Component gTextServiceJp = NULL;
static Component gTextServiceTCh = NULL;
static Component gTextServiceSCh = NULL;
static Component gTextServiceKr = NULL;

typedef struct 
	{
		char           *text; 
		int             textsize;
		int             textmax;

		TextRange      *hiliteRange;
		int             hilitesize;
		int             hilitemax;

		int             fixsize;
	} myAERecord;

typedef struct 
	{
		/* basic objects */
		EventTargetRef 		EventTarget;
		TSMDocumentID		docID;
		ComponentInstance 	ts;
	
		/* Input Methods */
		ComponentInstance 	tsJp;
		ComponentInstance 	tsTCh;
		ComponentInstance 	tsSCh;
		ComponentInstance 	tsKr;

		/* tracking Modes */
		CFStringRef   		CurrentModeString;
  		CFStringRef   		OldModeString;
  		CFStringRef   		CurrentScript;
		CFStringRef			OldScript;

		/* for event handling */
		EventHandlerRef 	EventHandler;
		WindowRef 			CandWind;
		myAERecord 			AERecord;
	} MacTSMInternals;

typedef MacTSMInternals* TSMHandle;

extern void MacImmSetCompositionString(char* text, long textSize);
extern void MacImmNotifyIMEComplete();

static OSErr OpenAppropriateService(TSMHandle internals, 
				    CFStringRef script)
{
	OSErr err = 1;

	TRACE("\n");

	if (internals->ts)
	{	
		/* Shut down current service */
		DeactivateTextService(internals->ts);
		TerminateTextService(internals->ts);
		internals->ts = NULL;
	}


	if (CFStringCompare(script, CFSTR("smJapanese"), 0) 
				== kCFCompareEqualTo)
	{
		if (!internals->tsJp)
			OpenTextService(internals->docID, gTextServiceJp, &(internals->tsJp));
		internals->ts = internals->tsJp;
	}
	else if (CFStringCompare(script, CFSTR("smTradChinese"), 0) 
				== kCFCompareEqualTo)
	{
		if (!internals->tsTCh)
			OpenTextService(internals->docID, gTextServiceTCh, &(internals->tsTCh));
		internals->ts = internals->tsTCh;
	}
	else if (CFStringCompare(script, CFSTR("smSimpChinese"), 0) 
				== kCFCompareEqualTo)
	{
		if (!internals->tsSCh)
			OpenTextService(internals->docID, gTextServiceSCh, &(internals->tsSCh));
		internals->ts = internals->tsSCh;
	}
	else if (CFStringCompare(script, CFSTR("smKorean"), 0) 
				== kCFCompareEqualTo)
	{
		if (!internals->tsKr)
			OpenTextService(internals->docID, gTextServiceKr, &(internals->tsKr));
		internals->ts = internals->tsKr;
	}
	else if (CFStringCompare(script, CFSTR("smRoman"), 0) 
				== kCFCompareEqualTo)
	{
		KeyScript(smKeyRoman);
		err = noErr;
	}

	if (internals->ts)
	{
		err = InitiateTextService(internals->ts);
		if (err != noErr)
			WARN("Failure to Initiate Text Service (%i)\n",err);
		else
		{	
			err = ActivateTextService(internals->ts);
			if (err != noErr)
			{
				TerminateTextService(internals->ts);
				WARN("Failure to Activate Text Service (%i)\n",err);
			}
		}

		if (err != noErr)
			internals->ts = NULL;
	}

	return err;
}


static Boolean SetInputModeCFS(TSMHandle internals, CFStringRef mode)
{
	OSStatus status;
	char ss[100];
	
	CFStringGetCString(mode,ss,100,kCFStringEncodingASCII);
	TRACE("Attempting to set mode to: %s\n", ss);

	if (!internals->ts)
		return false;
	
	status = SetTextServiceProperty(internals->ts,
					kTextServiceInputModePropertyTag, (SInt32)mode);

	if (status != noErr) 
	{
		ERR("can't set input mode (%ld)\n", status);
		return false;
	}
		
	return true;
}

static CFStringRef GetAquaDefaultInputModeCFS(CFStringRef theScript)
{
	CFDictionaryRef dict;
	CFArrayRef array = NULL;
	CFStringRef string = NULL;
	CFStringRef mode = NULL;

	CFPreferencesSynchronize(CFSTR("com.apple.HIToolbox"), 
			kCFPreferencesCurrentUser, kCFPreferencesCurrentHost);
	
	dict = CFPreferencesCopyValue(CFSTR("AppleDefaultInputMode"),
			CFSTR("com.apple.HIToolbox"),
			kCFPreferencesCurrentUser, kCFPreferencesCurrentHost);
	if (dict) 
	{
		array = CFDictionaryGetValue(dict, theScript);
		if (array)
		{
			string = CFArrayGetValueAtIndex(array, 0);
			mode = CFStringCreateCopy(NULL,string);
		}
		CFRelease(dict);
	}

	if (mode)
        {
            char ss[100];
            CFStringGetCString(mode,ss,100,kCFStringEncodingASCII);
            TRACE("aqua default input mode: %s\n", ss);
        }
	return mode;
}


static long GetAquaInputModeCFS(TSMHandle internals, 
				CFStringRef *out_mode, CFStringRef *out_script)
{
	CFStringRef string;
	long inIME = 0;
	char ss[100];

	CFPreferencesSynchronize(CFSTR("com.apple.HIToolbox"), 
		kCFPreferencesCurrentUser, kCFPreferencesCurrentHost);
	
	string = CFPreferencesCopyValue(CFSTR("AppleKeyboardScript"),
			CFSTR("com.apple.HIToolbox"),
			kCFPreferencesCurrentUser, kCFPreferencesCurrentHost);
	if (string) 
	{
		CFStringGetCString(string,ss,100,kCFStringEncodingASCII);
		TRACE("Aqua Script %s\n",ss);
		if (CFStringCompare(string, CFSTR("smRoman"), 0) == kCFCompareEqualTo)
			inIME = 0;
		else
			inIME = 1;

		if (inIME)
			*out_mode = GetAquaDefaultInputModeCFS(string);
		else
			*out_mode = CFStringCreateCopy(NULL, kTextServiceInputModeRoman);

		*out_script = string;
	}
	
	return inIME;
}

Boolean MacTSMRevertToPreviousMode(TSMHandle internals)
{
	if (CFStringCompare(internals->OldScript, internals->CurrentScript, 0) !=
				kCFCompareEqualTo)
	{
		char ss[100],ss2[100];
		CFStringRef temp = internals->CurrentScript;

		CFStringGetCString(temp,ss,100,kCFStringEncodingASCII);
		CFStringGetCString(internals->OldScript,ss2,100,kCFStringEncodingASCII);
		TRACE("Switching from %s to %s\n",ss,ss2);

		OpenAppropriateService(internals, internals->OldScript);

		internals->CurrentScript = internals->OldScript;
		internals->OldScript = temp;
	}

	if (CFStringCompare(internals->OldModeString, 
						internals->CurrentModeString,0) != kCFCompareEqualTo)
	{
		CFStringRef temp = internals->CurrentModeString;
		SetInputModeCFS(internals,internals->OldModeString);
		internals->CurrentModeString = internals->OldModeString;
		internals->OldModeString = temp;
	}

	return true;
}

Boolean MacTSMCheckModeAndIME(TSMHandle internals)
{
	Boolean inIME;
	CFStringRef mode = NULL;
	CFStringRef script = NULL;

	TRACE("\n");

	inIME = GetAquaInputModeCFS(internals, &mode, &script);

	if (!mode || !script)
		return false;

	if (CFStringCompare(script, internals->CurrentScript, 0) != 
			kCFCompareEqualTo)
	{
		OpenAppropriateService(internals, script);

		CFRelease(internals->OldScript);
		internals->OldScript = internals->CurrentScript;
		internals->CurrentScript = script;
	}
	else
		CFRelease(script);

	if (mode)
	{
		if (CFStringCompare(mode, internals->CurrentModeString,0)
			!= kCFCompareEqualTo)
		{
			CFRelease(internals->OldModeString);
			internals->OldModeString =
				internals->CurrentModeString;
			internals->CurrentModeString = mode;
			SetInputModeCFS(internals, internals->CurrentModeString);
		}
		else
			CFRelease(mode);
	}

	return inIME;
}

Boolean MacTSMEventCFS(TSMHandle internals, char theChar, unsigned long theCode,
		      unsigned long modifiers)
{
	OSStatus status;
	EventRef event = NULL;
	Boolean result;
	UInt32 Keyboard = LMGetKbdType();
	
	TRACE("(%02x'%c':%02lx:%04lx)\n", 
			(unsigned char)theChar, isprint(theChar) ? theChar : '.', 
			(long)theCode, (long)modifiers);

	if (!internals->ts)
		return false;

	status = CreateEvent(NULL, kEventClassKeyboard, kEventRawKeyDown, 0, 
						kEventAttributeUserEvent, &event);

	if (status == noErr)
		status = SetEventParameter(event, kEventParamKeyMacCharCodes, 
								typeChar, sizeof(char), &theChar);
		
	if (status == noErr)
		status = SetEventParameter(event, kEventParamKeyCode, typeUInt32, 
									sizeof(UInt32), &theCode);
		
	if (status == noErr)
		status = SetEventParameter(event, kEventParamKeyModifiers, typeUInt32, 
									sizeof(UInt32), &modifiers);
		
	if (status == noErr)
		status = SetEventParameter(event, kEventParamKeyboardType, typeUInt32,
								 sizeof(UInt32), &Keyboard);
		
	if (status == noErr) 
	{
		status = TextServiceEventRef(internals->ts, event);
		TRACE("Handled? %li\n",status);
		if (status > 0)
			result = true;
		else
			result = false;
	} 
	else 
	{
		ERR("can't send key to tsm (%ld)\n", status);
		result = false;
	}
	
	if (event)
		ReleaseEvent(event);

	if (result)
		while (CFRunLoopRunInMode(kCFRunLoopDefaultMode,0.0001,true)
				==kCFRunLoopRunHandledSource);

	return result;
}

TSMHandle MacInitTSMAwareApplication(void)
{
	OSStatus status;
	ScriptLanguageRecord scriptLanguageTCh = { smTradChinese, langTradChinese};
	ScriptLanguageRecord scriptLanguageSCh = { smSimpChinese, langSimpChinese};
	ScriptLanguageRecord scriptLanguageKr = { smKorean, langKorean};
	ScriptLanguageRecord scriptLanguageJp = { smJapanese, langJapanese};
	TSMHandle internals;
	
	TRACE("\n");

	internals = (TSMHandle)malloc(sizeof(MacTSMInternals));
	memset(internals,0,sizeof(MacTSMInternals));
	
	internals->EventTarget = GetApplicationEventTarget();

	memset(&(internals->AERecord),0,sizeof(myAERecord));

	/* Initialize Modes */
	internals->CurrentModeString = CFStringCreateCopy(NULL,
                                        kTextServiceInputModeRoman);
	internals->OldModeString = CFStringCreateCopy(NULL,
                                        kTextServiceInputModeRoman);
	internals->CurrentScript= CFStringCreateCopy(NULL,CFSTR("smRoman"));
	internals->OldScript= CFStringCreateCopy(NULL,CFSTR("smRoman"));

	status = GetDefaultInputMethodOfClass(&gTextServiceJp, &scriptLanguageJp, 
							kKeyboardInputMethodClass);
	if (status != noErr) 
		WARN("can't get Japanese im (%ld)\n", status);

	status = GetDefaultInputMethodOfClass(&gTextServiceTCh, &scriptLanguageTCh, 
							kKeyboardInputMethodClass);
	if (status != noErr) 
		WARN("can't get Traditional Chinese im (%ld)\n", status);

	status = GetDefaultInputMethodOfClass(&gTextServiceSCh, &scriptLanguageSCh, 
							kKeyboardInputMethodClass);
	if (status != noErr) 
		WARN("can't get Simplified Chinese im (%ld)\n", status);

	status = GetDefaultInputMethodOfClass(&gTextServiceKr, &scriptLanguageKr, 
							kKeyboardInputMethodClass);
	if (status != noErr) 
		WARN("can't get Korean im (%ld)\n", status);

	if (!gTextServiceTCh && !gTextServiceJp && !gTextServiceKr && !gTextServiceSCh)
	{
		WARN("Unable to create any text services\n");
		free(internals);
		return NULL;	
	}
	
	return internals;
}

void MacCloseTSMAwareApplication(TSMHandle internals)
{
	TRACE("\n");
	
	CFRelease(internals->CurrentModeString);
	CFRelease(internals->OldModeString);
	CFRelease(internals->CurrentScript);
	CFRelease(internals->OldScript);

	/* clean out the myAERecord */
	if (internals->AERecord.hiliteRange)
		free(internals->AERecord.hiliteRange);
	if (internals->AERecord.text)
		free(internals->AERecord.text);

	free(internals);
}


OSErr MacNewTSMDocument(TSMHandle internals, long refcon)
{
	OSErr err;
	InterfaceTypeList typeList;
	
	TRACE("\n");
	
	typeList[0] = kUnicodeDocumentInterfaceType;
	err = NewTSMDocument(1, typeList, &(internals->docID), refcon);
	TRACE("(%lx)\n", (long)(internals->docID));

	if (err != noErr)
	{
		ERR("can't create tsm doc (%d)\n", err);
		return err;
	}
		
	return noErr;
}

OSErr MacDeleteTSMDocument(TSMHandle internals)
{
	OSErr err;
	
	TRACE("\n");

	if (internals->ts)
	{
		DeactivateTextService(internals->ts);
		TerminateTextService(internals->ts);
	}

	if (internals->tsJp)
		CloseTextService(internals->docID, internals->tsJp);
	if (internals->tsTCh)
		CloseTextService(internals->docID, internals->tsTCh);
	if (internals->tsSCh)
		CloseTextService(internals->docID, internals->tsSCh);
	if (internals->tsKr)
		CloseTextService(internals->docID, internals->tsKr);

	err = DeleteTSMDocument(internals->docID);
	return err;
}

OSErr MacTSMFixTextService(TSMHandle internals)
{
	OSErr err;

	if (!internals->ts)
		return noErr;
	
	err = FixTextService(internals->ts);
	
	return err;
}

OSErr MacActivateTSMDocument(TSMHandle internals)
{
	OSErr err;

	err = ActivateTSMDocument(internals->docID);
	
	return err;
}


/**************************************
 * Event Handler Logic
 **************************************/
static unsigned long getText(EventRef event,myAERecord *theRec)
{
	unsigned long   ret;
	char           *theText = NULL;
	OSStatus        myErr;
	EventParamType  textType;

	textType = typeUnicodeText;
	myErr = GetEventParameter(event, kEventParamTextInputSendText, textType,
		   	NULL, 0, &ret, NULL);
	if (myErr != noErr)
		ret = 0;
	theRec->textsize = ret;

	if (ret != 0) 
	{
		if (ret >= theRec->textmax) 
		{
			free(theRec->text);
			theRec->text = (char *) malloc(sizeof(char) *
					 (theRec->textmax + 256));
			theRec->textmax += 256;
		}

		theText = theRec->text;

		myErr = GetEventParameter(event, kEventParamTextInputSendText, textType,
						NULL, ret, NULL, theText);
		if (myErr != noErr)
		theRec->textsize = ret = 0;
		*(theText + ret) = 0;

	} 
	else
 		*(theRec->text) = 0;

	if (TRACE_ON(mac_xim))
	{
		int i;

		TRACE("text(%ld): ", ret);
		for (i = 0; i < ret; i++) 
		{
			MESSAGE("%02x", (unsigned char)*(theText + i));
			MESSAGE("%s", (i % 2 == 0) ? "" : " ");
		}
		MESSAGE("\n");
	}

	return ret;
}


static void getTextRangeDesc(EventRef event,  myAERecord *theRec)
{
	static TextRangeArray *theArray = NULL;
	short           ret;
	int             i;
 	TextRange      *theRange;
	OSStatus        myErr;
	Size            actualSize;

	myErr = GetEventParameter(event, kEventParamTextInputSendHiliteRng,
				typeTextRangeArray, NULL, 0, (UInt32*)&actualSize, NULL);
	if (myErr == noErr) 
	{
		if (theArray == NULL)
		theArray = (TextRangeArray *)NewPtr(actualSize);
		else if (actualSize > GetPtrSize((Ptr)theArray)) 
		{
			DisposePtr((Ptr)theArray);
			theArray = (TextRangeArray *)NewPtr(actualSize);
		}

		myErr = GetEventParameter(event, kEventParamTextInputSendHiliteRng,
				typeTextRangeArray, NULL, actualSize, NULL, theArray);
		if (myErr != noErr)
			return;
  
		if (TRACE_ON(mac_xim))
		{
			TRACE("hilite(%d): ", theArray->fNumOfRanges);
			for (i = 0; i < theArray->fNumOfRanges; i++)
			MESSAGE("%s(%ld-%ld:%d)", i == 0 ? "" : ", ",
			theArray->fRange[i].fStart, theArray->fRange[i].fEnd,
			theArray->fRange[i].fHiliteStyle);
			MESSAGE("\n");
		}

		ret = theRec->hilitesize = theArray->fNumOfRanges;
		if (ret == 0) 
	    	return;
	} 
	else 
	{
		if (theArray != NULL)
			return;
		myErr = GetEventParameter(event, kEventParamTextInputSendUpdateRng,
					typeTextRangeArray, NULL, 0, (UInt32*)&actualSize, NULL);
		if (myErr != noErr)
			return;
		theArray = (TextRangeArray *)NewPtr(actualSize);
		myErr = GetEventParameter(event, kEventParamTextInputSendUpdateRng,
					typeTextRangeArray, NULL, actualSize, NULL, theArray);
		if (myErr != noErr)
			return;
		ret = theRec->hilitesize = theArray->fNumOfRanges;
		if (ret != 2)
			return;

		if (TRACE_ON(mac_xim))
		{
			TRACE("update(%d): ", theRec->hilitesize);
			for (i = 0; i < theRec->hilitesize; i++)
				MESSAGE("%s(%ld-%ld)", i == 0 ? "" : ", ",
					theArray->fRange[i].fStart, theArray->fRange[i].fEnd);
			MESSAGE("\n");
		}

		theArray->fRange[0].fStart = theArray->fRange[1].fStart;
		theArray->fRange[0].fEnd = theArray->fRange[1].fEnd;
		theArray->fRange[0].fHiliteStyle = 4;
		theArray->fRange[1].fStart = theArray->fRange[1].fEnd;
		theArray->fRange[1].fHiliteStyle = 1;
	}

	if (ret > theRec->hilitemax) 
	{
		free(theRec->hiliteRange);
		theRec->hiliteRange = (TextRange *) malloc(sizeof(TextRange) * ret);
		theRec->hilitemax = ret;
	}

	theRange = theRec->hiliteRange;
	for (i = 0; i < ret; i++) 
	{
		(theRange + i)->fStart = theArray->fRange[i].fStart;
		(theRange + i)->fEnd = theArray->fRange[i].fEnd;
		(theRange + i)->fHiliteStyle = theArray->fRange[i].fHiliteStyle;
	}
}

static OSStatus UpAcInputArea(EventRef event, TSMHandle internals)
{
	OSStatus        myErr;
	long            fixLength;
	long            textSize;
	long            refCon;
	myAERecord     *theRec;
  
	theRec = &(internals->AERecord);

	myErr = GetEventParameter(event, kEventParamTextInputSendRefCon,
				typeLongInteger, NULL, sizeof(long), NULL, &refCon);

	if (myErr == noErr) 
	{
		textSize = getText(event,theRec);
		TRACE("MacIM:AEUpdateEvent:textSize=%ld\n", textSize);
	}

	if (myErr != noErr) 
	{
		return myErr;
	}

	myErr = GetEventParameter(event, kEventParamTextInputSendFixLen,
				typeLongInteger, NULL, sizeof(long), NULL, &fixLength);
	if (myErr == noErr) 
	{
		theRec->fixsize = fixLength;
		TRACE("HERE fix(%d)\n", theRec->fixsize);
	} 
	else 
		return myErr;

	if (textSize >= fixLength) 
		getTextRangeDesc(event,theRec);

	if (textSize > 0)
	{
		MacImmSetCompositionString(theRec->text,textSize);

		if (theRec->fixsize == textSize)
			MacImmNotifyIMEComplete();
	}

	return noErr;
}

/* Apple Event Off2Pos */ 
static OSStatus Off2P(EventRef event)
{
	OSStatus        myErr;
	long            refCon;
	BitMap          screenBits;
	Point           pos = { 0, 0 };

  /* extern void     getCurrTextPos(long, Point *); */
  
	myErr = GetEventParameter(event, kEventParamTextInputSendRefCon,
				typeLongInteger, NULL, sizeof(long), NULL, &refCon);

/*
	if (myErr == noErr)
		getCurrTextPos(refCon, &pos);
*/
    
	if (pos.v == 0 && pos.h == 0) 
	{
		GetQDGlobalsScreenBits(&screenBits);
  
		pos.h = screenBits.bounds.right - kCandWindWidth - 4;
		pos.v = screenBits.bounds.top + kMenuBarHeight + 4;
	}

	myErr = SetEventParameter(event, kEventParamTextInputReplyPoint,
					typeQDPoint, sizeof(Point), &pos);
	return myErr;
}

static void UpdateCandWindow(TSMHandle internals)
{
	/* 
	 * Mac Documentation says that this call is Deprecated
	 */
	QDFlushPortBuffer(GetWindowPort(internals->CandWind), NULL);
}

static pascal OSStatus MacEventHandler(EventHandlerCallRef nextHandler, 
					EventRef event, void *userData)
{
#pragma unused (nextHandler)

	OSStatus        myErr, result = eventNotHandledErr;
	UInt32          class, kind;
	WindowRef       wind;
	TSMHandle		internals = (TSMHandle)userData;

	class = GetEventClass(event);
	kind = GetEventKind(event);
  
	TRACE("(%c%c%c%c:%ld)\n", (char)(class >> 24) & 0xff, 
		(char)(class >> 16) & 0xff, (char)(class >> 8) & 0xff, 
		(char)class & 0xff, kind);
    
	switch (class) 
	{
		case kEventClassTextInput:
				switch (kind) 
				{
					case kEventTextInputUpdateActiveInputArea:	
						TRACE("kEventTextInputUpdateActiveInputArea\n");
						if (internals->CandWind)
							UpdateCandWindow(internals);
						myErr = UpAcInputArea(event,internals);
						if (myErr == noErr)
							result = noErr;
						break;
					case kEventTextInputOffsetToPos:
						TRACE("kEventTextInputOffsetToPos\n");
						myErr = Off2P(event);
						if (myErr == noErr)
							result = noErr;
						break;
				}
				break;
		case kEventClassWindow:
			TRACE("kEventClassWindow\n");
			myErr = GetEventParameter(event, kEventParamDirectObject,
						typeWindowRef, NULL, sizeof(WindowRef), NULL, &wind);
			if (myErr != noErr)
				break;
			switch (kind) 
			{
				case kEventWindowShown:
					TRACE("kEventWindowShown\n");
					if (internals->CandWind == NULL) 
					{
						internals->CandWind = wind;
						UpdateCandWindow(internals);
					}
					break;
				case kEventWindowHidden:
					TRACE("kEventWindowHidden\n");
					if (internals->CandWind == wind)
						internals->CandWind = NULL;
					break;
			}
			break;
 	}

	return result;
}

long MacInstallAppEventHandler(TSMHandle internals)
{
	OSStatus        myErr;
	EventTypeSpec   eventType[] = {
			{ kEventClassTextInput, kEventTextInputUpdateActiveInputArea },/*1*/
			{ kEventClassTextInput, kEventTextInputOffsetToPos },	/* 3 */
			{ kEventClassWindow, kEventWindowShown },	/* 24 */
			{ kEventClassWindow, kEventWindowHidden },	/* 25 */
	};
	
	TRACE("Setting up Event Handler\n");
	myErr = InstallApplicationEventHandler(NewEventHandlerUPP(MacEventHandler), 
				GetEventTypeCount(eventType), eventType, internals, 
				&(internals->EventHandler));

	if (myErr != noErr) 
	{
		ERR("can't install app event handler (%ld)\n", myErr);
		return myErr;
	}

	return noErr;
}

void MacRemoveAppEventHandler(TSMHandle internals)
{
	OSStatus        myErr;
	
	myErr = RemoveEventHandler(internals->EventHandler);
}

#endif /* __APPLE__ */
